unit RTSP;

{ RTSP Component for Delphi (c)2006 by Paul TOTH <tothpaul@free.fr> }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}


interface

{-$define log}

uses
 Windows, Messages, Winsock, Classes, SysUtils;

const
 VERSION='0.2';
 USER_AGENT='TVInside ('+VERSION+')';

type
 TEvent=(evError,evWSError,evSend,evRecv,evCtrl,evDone);

 TEventMethod=procedure(Sender:TObject; Event:TEvent; Value:integer) of object;
 TDecodeEvent=procedure(Sender:TObject; Size:integer; Data:pointer) of object;

 TRTSPClient=class
 private
  fHWnd      :integer;
  fHandle    :integer;
  fHost      :string;
  fURL       :string;
  fPort      :integer;
  fActive    :boolean;
  fPlaying   :boolean;
  fRTPThread :THandle;
  fRTSPThread:THandle;
  fRemote    :TSockAddr;
  fControl   :TSockAddr;
  fCSeq      :integer;
  fRequest   :string;
  fReply     :string;
  fHeader    :string;
  fData      :string;
  fSession   :string;
  EOnEvent   :TEventMethod;
  EOnDecode  :TDecodeEvent;
  procedure RTPThread;
  procedure RTSPThread;
  function UDPSocket(Port:integer):integer;
  function Connect:boolean;
  function SendRequest:boolean;
  function ReadReply:integer;
  function ExecRequest(const Request:string; Parms:string=''):integer;
  function ReadMore:boolean;
  function IntegerAt(Offset:integer):integer;
  function getHeader(const Tag:string):string;
  function GetServerPort:boolean;
  function BuildReport:string;
 protected
  procedure WndProc(var Msg:TMessage);
  procedure DoEvent(Event:TEvent; Value:integer); virtual;
  procedure DoDecode(Size:integer; Data:pointer); virtual;
 public
  constructor Create;
  destructor Destroy; override;
  function Open(const URL:string; Port:integer=554):boolean;
  procedure Stop;
  property Request:string read fRequest;
  property Header:string read fHeader;
  property Data:string read fData;
  property Playing:boolean read fPlaying;
  property OnEvent:TEventMethod read EOnEvent write EOnEvent;
  property OnDecode:TDecodeEvent read EOnDecode write EOnDecode;
 end;

implementation

const
 RTSP_ERROR = WM_USER+1;
 RTSP_SEND  = WM_USER+2;
 RTSP_RECV  = WM_USER+3;
 RTSP_DATA  = WM_USER+4;
 RTSP_CTRL  = WM_USER+5;
 RTSP_DONE  = WM_USER+6;
 RTSP_WSERROR=WM_USER+7;

 RTSP_HOST_NOT_FOUND = 1;
 RTSP_WINSOCK_ERROR  = 2;
 RTSP_PROTOCOL_ERROR = 3;

type
 TRTPHeader=packed record
 // HEADER 12 bytes
  Version  : byte;      // $80  : Version = 2
  Payload  : byte;      // 33   : MPEG2TS
  Sequence : word;
  Timestamp: cardinal;
  SSRC     : cardinal;
 // MPEG2 PayLoad...
 end;

 TRTCPReport=packed record
 // HEADER (8 bytes)
  h1Version : byte;     // $81 : Version = 2 , Report Count =1
  h1Code    : byte;     // 201
  h1Length  : word;     //   7 : (32/4)-1
  h1SSRC    : cardinal; //
 // RECEIVER REPORT (24 bytes)
  rrSSRC    : cardinal;
  rrLost    : cardinal;
  rrHighest : word;
  rrSequence: word;     // last Recevied sequence
  rrJitter  : cardinal;
  rrLastSR  : cardinal;
  rrDelay   : cardinal;
 // HEADER (8 bytes)
  h2Version : byte;     // $81 : Version = 2 , no Padding, Report Count =1
  h2Code    : byte;     // 202
  h2Length  : word;     //
  h2SSRC    : cardinal; //
 // Source Description
  ssCode    : byte;     // 1
  ssLength  : byte;     // Length(ssName)
  ssName    : char;     // var length string
 // Pad  
 end;

// Thread entry points
function StartRTP(Sender:TRTSPClient):integer; stdcall;
begin
 try
  Sender.RTPThread;
  Result:=0;
 except
  Result:=-1;
 end;
end;

function StartRTSP(Sender:TRTSPClient):integer; stdcall;
begin
 try
  Sender.RTSPThread;
  Result:=0;
 except
  Result:=-1;
 end;
end;

// Wait or Kill a Thread
procedure WaitOrKill(var Thread:THandle; Time:cardinal);
begin
 if Thread=0 then exit;
 if WaitForSingleObject(Thread,Time)=WAIT_TIMEOUT then begin
  TerminateThread(Thread,0);
  CloseHandle(Thread);
  Thread:=0;
 end;
end;

// Extract Host, Path and Port from an URL
function DecodeURL(url:string; var Host,Path:string; var Port:integer):boolean;
var
 i,e:integer;
begin
 Result:=False;
// "rtsp://"
 i:=pos('://',url);
 if i>0 then begin
  if copy(url,1,i-1)<>'rtsp' then exit;
  Delete(url,1,i+2);
 end;
// host/path
 i:=pos('/',url);
 if i=0 then begin
  Host:=url;
  Path:='/';
 end else begin
  Host:=copy(url,1,i-1);
  Path:=copy(url,i,MaxInt);
 end;
// host:port
 i:=pos(':',Host);
 if i=0 then
  Port:=554
 else begin
  val(copy(Host,i+1,MaxInt),Port,e);
  if e<>0 then exit;
  SetLength(Host,i-1);
 end;
 Result:=True;
end;

// Get local HostName
function GetHostName:string;
var
 host:array[0..80] of char;
begin
 if Winsock.gethostname(host,sizeof(host))=0 then
  Result:=host
 else
  Result:='localhost';
end;

// Retreive Host IP Address
function INetAddr(Host:string):integer;
var
 pHost:PChar;
 HostEnt:PHostEnt;
begin
 if Host='' then begin
  Result:=INADDR_NONE;
 end else begin
  pHost:=PChar(Host);
  Result:=inet_addr(pHost);
  if Result=INADDR_NONE then begin
   HostEnt:=gethostbyname(pHost);
   if HostEnt<>nil then Result:=integer(pointer(HostEnt^.h_addr^)^);
  end;
 end;
end;

//----------------------------------------------------------------------------//

constructor TRTSPClient.Create;
begin
 fHWnd:=AllocateHWnd(WndProc);
end;

destructor TRTSPClient.Destroy;
begin
 Stop;
 WaitOrKill(fRTPThread,50);
 WaitOrKill(fRTSPThread,50);
 DeallocateHWnd(fHWnd);
 inherited;
end;

// Windows way to synchronize threads
procedure TRTSPClient.WndProc(var Msg:TMessage);
begin
 case Msg.Msg of
  RTSP_ERROR  : DoEvent(evError,Msg.lParam);
  RTSP_WSERROR: DoEvent(evWSError,Msg.lParam);
  RTSP_SEND   : DoEvent(evSend,0);
  RTSP_RECV   : DoEvent(evRecv,0);
  RTSP_DATA   : DoDecode(Msg.wParam,pointer(msg.lParam));
  RTSP_CTRL   : DoEvent(evCtrl,Msg.LParam);
  RTSP_DONE   : DoEvent(evDone,Msg.lParam);
 end;
end;

procedure TRTSPClient.DoEvent(Event:TEvent; Value:integer);
begin
 if Assigned(EOnEvent) then EOnEvent(Self,Event,Value);
end;

procedure TRTSPClient.DoDecode(Size:integer; Data:pointer);
begin
 if Assigned(EOnDecode) then EOnDecode(Self,Size,Data);
end;

// Stop RTP stream by releasing self suspended RTSP
procedure TRTSPClient.Stop;
begin
 if fPlaying then ResumeThread(fRTSPThread) else fActive:=False;
end;

// Start downloading an RTSP resource
function TRTSPClient.Open(const URL:string; Port:integer=554):boolean;
var
 id:cardinal;
begin
 Result:=False;
 if not DecodeURL(URL,fHost,fURL,fPort) then exit;
 IsMultiThread:=True;
 fActive:=True;
 fRTPThread:=CreateThread(nil,0,@StartRTP,Self,0,id);
 Result:=True;
end;

// Open a UDP socket
function TRTSPClient.UDPSocket(Port:integer):integer;
var
 sa:TSockAddr;
begin
 Result:=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
 if Result=INVALID_SOCKET then begin
  PostMessage(fHWnd,RTSP_WSERROR,0,WSAGetLastError);
 end else begin
  FillChar(sa,SizeOf(sa),0);
  sa.sin_family:=AF_INET;
  sa.sin_port:=htons(Port);
  if bind(Result,sa,SizeOf(sa))<>0 then begin
   SendMessage(fHWnd,RTSP_WSERROR,0,WSAGetLastError);
   closesocket(Result);
   Result:=INVALID_SOCKET;
  end;
 end;
end;

// Build a RTCP report data block
function TRTSPClient.BuildReport:string;
var
 report:^TRTCPReport;
 l,rsize:integer;
begin
// CName
 Result:='TVInside@'+GetHostName;
 l:=length(Result);
// Report size
 rsize:=SizeOf(TRTCPReport)+l;
// Align size
 rsize:=(rsize+3)and(not 3);
// s will store the Report
 SetLength(Result,rsize);
 report:=@Result[1];
// move CName to the correct place
 move(Result[1],report.ssName,l);
// blank the header
 FillChar(report^,SizeOf(TRTCPReport)-1,0);
// fill header
 report.h1Version:=$81;
 report.h1Code:=201;
 report.h1Length:=swap(7); // (32/4)-1
 report.h1SSRC:=GetTickCount;
 report.h2Version:=$81;
 report.h2Code:=202;
 report.h2Length:=swap(((rsize-32) div 4)-1);
 report.h2SSRC:=report.h1SSRC;
 report.ssCode:=1;
 report.ssLength:=l;
end;

// Listening thread
procedure TRTSPClient.RTPThread;
var
 rtp    :integer;
 rtcp   :integer;
 Buffer :array[0..1500] of byte;
 Header :TRTPHeader absolute Buffer;
 From   :TSockAddr;
 Size   :integer;
 i      :integer;
 report :^TRTCPReport;
 s      :string;
 count  :integer;
 total  :integer;
begin
 total:=0;
 try
  rtp:=UDPSocket(32768); // we need a rtp listening socket
  if rtp=INVALID_SOCKET then exit;
  try
   rtcp:=UDPSocket(32769); // and a rtcp report socket
   if rtcp=INVALID_SOCKET then exit;
   try
    count:=0;
    s:=BuildReport; // we use a string to hold the report
    report:=@s[1];
   // now we can request the stream
    fRTSPThread:=CreateThread(nil,0,@StartRTSP,Self,0,cardinal(i));
    while fActive do begin
     FillChar(From,SizeOf(From),0);
     From.sin_family:=AF_INET;
     Size:=SizeOf(From);
     i:=recvfrom(rtp,Buffer,SizeOf(Buffer),0,From,Size);
     if i=0 then exit;
     if i<0 then begin
      SendMessage(fHWnd,RTSP_WSERROR,0,WSAGetLastError);
     end else
     if i>12 then begin // need at leat 12 bytes for header
      if (Header.Version=$80)and(Header.Payload=33) then begin
      // RTP version 2, with MPEG2TS payload
       dec(i,12);
       inc(total,i);
       SendMessage(fHWnd,RTSP_DATA,i,integer(@Buffer[12]));
       // RTCP
       inc(count);
       if count=1500 then begin // experimental
        count:=0;
       // like live555 in VLC, we just fill the Sequence entry
        report.rrSequence:=Header.Sequence;
        sendto(rtcp,s[1],length(s),0,fControl,SizeOf(fControl));
        SendMessage(fHWnd,RTSP_CTRL,0,swap(Header.Sequence));
       end;
      end;
     end;
    end;
   finally
    closesocket(rtcp);
   end;
  finally
   closesocket(rtp);
  end;
 finally
  SendMessage(fHWnd,RTSP_DONE,0,total);
  CloseHandle(fRTPThread);
  fRTPThread:=0;
 {$ifdef log}WriteLn('RTPTrhead done');{$ENDIF}
 end;
end;

// Try to connect to the RTSP host
function TRTSPClient.Connect:boolean;
begin
 Result:=False;
 FillChar(fRemote,SizeOf(fRemote),0);
 fRemote.sin_family:=AF_INET;
 fRemote.sin_port:=htons(fPort);
 fRemote.sin_addr.S_addr:=INetAddr(fHost);
 if fRemote.sin_addr.S_addr=INADDR_NONE then begin
  SendMessage(fHwnd,RTSP_ERROR,0,RTSP_HOST_NOT_FOUND);
  exit;
 end;
 fHandle:=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
 if fHandle=INVALID_SOCKET then begin
  SendMessage(fHwnd,RTSP_WSERROR,0,WSAGetLastError);
  exit;
 end;
 if Winsock.connect(fHandle,fRemote,SizeOf(fRemote))<>0 then begin
  SendMessage(fHwnd,RTSP_WSERROR,0,WSAGetLastError);
  CloseSocket(fHandle);
  fHandle:=INVALID_SOCKET;
  exit;
 end;
 Result:=True;
end;

// RTSP thread
procedure TRTSPClient.RTSPThread;
begin
 try
  if not Connect then exit;
  try
   if ExecRequest('SETUP','Transport: RTP/AVP;unicast;client_port=32768-32769'#13#10)<>200 then exit;
   if not GetServerPort then exit;
   if ExecRequest('PLAY')<>200 then exit;
   fPlaying:=True;
   if fActive then begin
   // now we just wait for a Resume to stop the stream
    SuspendThread(fRTSPThread);
   // stop RTP listening
    fActive:=False;
   // and wait for the thread to finish or recvfrom() will fall in a never ending loop
    WaitForSingleObject(fRTPThread,100);
   end;
   if ExecRequest('TEARDOWN')<>200 then exit;
  finally
   fPlaying:=False;
   closesocket(fHandle);
   fHandle:=INVALID_SOCKET;
  end;
 finally
  CloseHandle(fRTSPThread);
  fRTSPThread:=0;
 {$ifdef log}WriteLn('RTSPTrhead done');{$ENDIF}
 end;
end;

function TRTSPClient.ExecRequest(const Request:string; Parms:string=''):integer;
begin
 if (fSession<>'') then begin
  Parms:=Parms+'Session: '+fSession+#13#10;
  if Request='TEARDOWN' then fSession:='';
 end;
 inc(fCSeq);
 fRequest:=Request+' rtsp://'+fHost+fURL+' RTSP/1.0'#13#10+
          'CSeq: '+IntToStr(fCSeq)+#13#10+
           Parms+
          'User-Agent: '+USER_AGENT+#13#10+
          #13#10;
 if SendRequest then Result:=ReadReply else Result:=-1;
end;

function TRTSPClient.SendRequest:boolean;
var
 i,l,x:integer;
begin
 Result:=False;
 l:=Length(fRequest);
 x:=1;
 while l>0 do begin
  i:=send(fHandle,fRequest[x],l,0);
  if i<=0 then begin
   SendMessage(fHWnd,RTSP_WSERROR,0,WSAGetLastError);
   exit;
  end;
  inc(x,i);
  dec(l,i);
 end;
 SendMessage(fHWnd,RTSP_SEND,0,0);
 Result:=True;
end;

// retrieve an integer at a specific position in the header
function TRTSPClient.IntegerAt(Offset:integer):integer;
var
 num:char;
begin
 Result:=0;
 while (Offset<=Length(fHeader)) do begin
  num:=fHeader[Offset];
  if num in ['0'..'9'] then
   Result:=10*Result+ord(num)-ord('0')
  else
   exit;
  inc(Offset);
 end;
end;

// retreive a header field
function TRTSPClient.getHeader(const Tag:string):string;
var
 i,j:integer;
begin
 Result:='';
 i:=pos(#13#10+Tag+':',fHeader);
 if i=0 then exit;
 inc(i,2+Length(Tag)+1);
 while (i<=Length(fHeader))and(fHeader[i]=' ') do inc(i);
 j:=i;
 while (j<=Length(fHeader))and(fHeader[j]<>#13) do inc(j);
 Result:=copy(fHeader,i,j-i);
end;

// retreive the server RTCP port from the header
function TRTSPClient.GetServerPort:boolean;
var
 s:string;
 i,p:integer;
begin
// Transport: RTP/AVP;unicast;mode=play;destination=192.168.0.2;client_port=4072-4073;server_port=32768-32769
 Result:=False;
 s:=getHeader('Transport');
 i:=pos('server_port=',s);
 if i=0 then exit;
 Delete(s,1,i+11);
 i:=pos('-',s);
 if i=0 then exit;
 Delete(s,1,i);
 Val(s,p,i);
 Result:=i=0;
 fControl:=fRemote;
 fControl.sin_port:=htons(p);
end;

// generic RTSP reply receiver
function TRTSPClient.ReadReply:integer;
var
 s  :string;
 i  :integer;
 len:integer;
begin
 Result:=-1;
 while fActive do begin
  i:=pos(#13#10#13#10,fReply);
  if i>0 then begin
   fHeader:=copy(fReply,1,i+1);
   Delete(fReply,1,i+3);
   if copy(fHeader,1,8)<>'RTSP/1.0' then begin
    SendMessage(fHWnd,RTSP_ERROR,0,RTSP_PROTOCOL_ERROR);
    exit;
   end;
   s:=getHeader('Session');
   if fSession='' then
    fSession:=s
   else begin
    if s<>fSession then begin
     SendMessage(fHWnd,RTSP_ERROR,0,RTSP_PROTOCOL_ERROR);
     exit;
    end;
   end;
   i:=pos('Content-Length: ',fHeader);
   if i>0 then begin
    len:=IntegerAt(i+16);
    while fActive do begin
     if Length(fReply)>=len then break;
     if ReadMore=False then exit;
    end;
    fData:=copy(fReply,1,len);
    Delete(fReply,1,len);
   end;
   SendMessage(fHWnd,RTSP_RECV,0,0);
   Result:=IntegerAt(10);
   exit;
  end;
  if ReadMore=False then exit;
 end;
end;

// fill receive buffer
function TRTSPClient.ReadMore:boolean;
var
 i,l:integer;
 s:string;
begin
 l:=1024;
 SetLength(s,l);
 i:=recv(fHandle,s[1],l,0);
 if i<=0 then begin
  SendMessage(fHWnd,RTSP_WSERROR,0,WSAGetLastError);
  Result:=False;
  exit;
 end;
 fReply:=fReply+copy(s,1,i);
 Result:=True;
end;

procedure WSAStartup;
var
 WSA:TWSAData;
begin
 winsock.WSAStartup($101,wsa);
end;

initialization
 WSAStartup;
{$ifdef log}AllocConsole;{$endif}
finalization
 WSACleanup;
end.

